function [ W,H,Rs1,Rs2,Rs3,Rs4,Rs5,fobjec,ResF] = ...
  Lea_Algorithm_T_D2(X,tol,maxiter,noc,Cof_Lambda,U,V,Cof_Alpha,Outdisplay)
% function Lea_Algorithm_T_D2 performs the ADMM algortihm in k iterations
 % for the dictionary learning problem in the Tensor form, where the
 % dictionary D belongs to the set D_2
   
% Usage:
%[ W,H,Rs1,Rs2,Rs3,Rs4,Rs5,fobjec,ResF] = Lea_Algorithm_T_D2...
%                 (X,tol,maxiter,noc,Cof_Lambda,U,V,Cof_Alpha)
%
% Input:
% X: Training Tensor of size: patch_size(1) by t by patch_size(2),
% tol: The tolerance for the ADMM convergance
% maxiter: total number of ADMM iterations
% noc: The number of dictionary elements,
% Cof_Lambda: the regularization parameter Lambda in the dictionary
% learning problem formulation
% U,V: The initial value for the auxiliary variables in the ADMM algorithm
% Cof_Alpha: The agumented lagrangian parameter
% Outdisplay: The display option on the screen, when set to 0 no display 
% on the screen, when set to 1 printes progress of the algorithm
% default = 0,


% Output:
% W: The dictionary tensor for the corresponding lambda of size:
% patch_size(1) by noc by patch_size(2)
% H: The representation tensor for the corresponding lambda of size:
% noc by t by patch_size(2)
% fobjec: The objective function value for each iteration of the ADMM
% algorithm: 1/2||Y_D*H||^2_F+\lambda||H||_{sum}
% ResF: The residual value for each iteration of the ADMM
% algorithm: ||Y_D*H||_F
% Rs1,Rs2,Rs3,Rs4,Rs5: The stopping critera values at each iteration 
% of the ADMM algorithm

% Copyright (C) Sara Soltani and Technical University of Denmark, 
% August 2015, ssol@dtu.dk, sarahsoltani@gmail.com
             
% initialization
[m,n,p] = size(X);


Lag_PI = zeros(noc,n,p);
Lag_Delta = zeros(m,noc,p);


H = V;

%%
fobjec=zeros(maxiter,1);
Rs1=zeros(maxiter,1);
Rs2=zeros(maxiter,1);
Rs3=zeros(maxiter,1);
Rs4=zeros(maxiter,1);
Rs5=zeros(maxiter,1);
ResF=zeros(maxiter,1);
%%

for k=1:maxiter
    
    % W update

    W=U-1/Cof_Alpha*Lag_Delta;
    
    %  norm 2 approach doing dykstra projections
     
    for i=1:noc
        [ pv ] = dykstra(reshape(squeeze(W(:,i,:)),m*p,1) ); 
        W(:,i,:) = reshape(pv,m,1,p);
    end
    
    % V update
    V = V_update(H,Cof_Alpha,X,Lag_PI,U,p,noc);
    
    % H update
    H = soft_thresh(Cof_Alpha,Cof_Lambda,V,Lag_PI,noc,n,p);

    % U update
    U = U_update(W,Cof_Alpha,X,Lag_Delta,V,p,noc);

    % Lagrangian Variable update
    Lag_Delta = Lag_Delta + Cof_Alpha*(W-U);
    Lag_PI = Lag_PI + Cof_Alpha*(H-V);


%%
    Rs1(k)=norm(W(:)-U(:),'inf')/max(1,norm(W(:),'inf'));
    Rs2(k) = norm(H(:)-V(:),'inf')/max(1,norm(H(:),'inf'));
    Rs=zeros(noc,1); 

    for i=1:noc
        tmp=squeeze(W(:,i,:));
         Rs(i,1)=norm(tmp(:),2);
     
    end
   

    % Dual Resudials
    Comp = X-tprod(W,H);

    tempp0=tprod(tran(W),Comp);
    tempp1=tprod(Comp,tran(H));
    Rs4(k) = norm(tempp0(:)+Lag_PI(:),'inf')/max(1,norm(tempp0(:),'inf'));
    Rs5(k) = norm(tempp1(:)+Lag_Delta(:),'inf')/...
        max(1,norm(tempp1(:),'inf'));
    
    Rs3(k)=max(abs(sqrt(m*p)-Rs)); 
    ResF(k)=fronorm(Comp);
    fobjec(k)=1/2*(ResF(k)^2)+Cof_Lambda*norm(H(:),1);
    if (Rs1(k)<= tol && Rs2(k)<=tol && Rs3(k)<=tol && Rs5(k)<=tol && ...
            Rs4(k)<=0.05 && fobjec(k)<=fobjec(1))
        if Outdisplay==1
           disp('Algorithm is converged. The tolerance reached') 
        end
    break
    end
    if Outdisplay==1
       disp(['The ADMM iteration number =', num2str(k)])
       disp(['The residual at iteration ', num2str(k),...
        ' i.e., ||Y-D*H||_F = ',num2str(ResF(k))])
    end
    
end

end
%%
function H = soft_thresh(Cof_Alpha,Cof_Lambda,V,Lag_PI,noc,n,p)


th = Cof_Lambda/(2*Cof_Alpha);   
x = zeros(noc*n*p,1);

B = (V - Lag_PI./Cof_Alpha); 
B = reshape(B,noc*n*p,1);

k = find(B > th);
x(k) = B(k)-th;

k = find(abs(B) <= th);
x(k) = 0;

k = find (B < -th);
 x(k) = 0; 

H = reshape(x,noc,n,p);

end
%% 
function V = V_update(H,Cof_Alpha,X,Lag_PI,U,p,noc)

V_hat=zeros(size(Lag_PI));

H_hat=fft(H,[],3);
U_hat=fft(U,[],3);
UT_hat=fft(tran(U),[],3);
PI_hat=fft(Lag_PI,[],3);
X_hat=fft(X,[],3);

for i=1:p   
    V_hat(:,:,i) = (UT_hat(:,:,i)*U_hat(:,:,i) + Cof_Alpha*eye(noc,noc))...
     \(UT_hat(:,:,i)*X_hat(:,:,i)+PI_hat(:,:,i) + Cof_Alpha*H_hat(:,:,i));
end
V=ifft(V_hat,[],3);
end

function U=U_update(W,Cof_Alpha,X,Lag_Delta,V,p,noc)

U_hat=zeros(size(Lag_Delta));

W_hat=fft(W,[],3);

V_hat=fft(V,[],3);
VT_hat=fft(tran(V),[],3);
Delta_hat=fft(Lag_Delta,[],3);
X_hat=fft(X,[],3);


for i=1:p
    U_hat(:,:,i) = (X_hat(:,:,i)*VT_hat(:,:,i)+Cof_Alpha*W_hat(:,:,i)+...
    Delta_hat(:,:,i))/(V_hat(:,:,i)*VT_hat(:,:,i)+Cof_Alpha*eye(noc,noc));
end
U=ifft(U_hat,[],3);

end
